import json
from bs4 import BeautifulSoup
import time
import sys
from datetime import datetime, timezone
import email.utils as eut
import logging
from catswalk.scraping.types.type_response import Response, ResponseHtml, ResponseJson
import requests

logger = logging.getLogger()


class CatsWalkRequestError(Exception):
    pass


class CWRequest:
    DEFAULT_TIMEOUT = (15.0, 20.5)

    def __init__(self, verify=True, timeout=DEFAULT_TIMEOUT):
        """[summary]

        Args:
            verify (bool, optional): [description]. Defaults to True.
        """
        self.verify = verify
        self.timeout = timeout
        self.session = requests.Session()
        self.session.verify = self.verify
        self.timeout = self.timeout

    def __enter__(self):
        """

        :return:
        """
        self.print_global_info()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.close()

    @classmethod
    def parse_http_date(s):
        """[summary]

        Args:
            s ([type]): [description]

        Returns:
            [type]: [description]
        """
        return datetime(*eut.parsedate(s)[:6], tzinfo=timezone.utc)

    def close(self):
        """[Close session]
        """
        self.session.close()

    def get_cookie(self, key: str):
        """[Get cookie by key]

        Args:
            key ([type]): [description]

        Returns:
            [type]: [description]
        """
        return self.session.cookies.get(key)

    def get_cookies(self):
        """[Get cookies]

        Returns:
            [type]: [description]
        """
        return self.session.cookies.get_dict()


    def __mk_result(self, ret, response_content_type: str):
        """[Wrap response]

        Args:
            ret ([type]): [description]
            response_content_type (str): [description]

        Returns:
            [type]: [description]
        """
        if response_content_type == "html":
            soup = BeautifulSoup(ret.content, features="html.parser")
            return ResponseHtml(ret.headers, soup, ret)
        elif response_content_type == "json":
            soup = BeautifulSoup(ret.content, features="html.parser")
            return ResponseJson(ret.headers, json.loads(str(soup)), ret)
        else:
            return Response(ret.headers, ret.content, ret)


    def __check_status_code(self, url, status_code):
        """[Check http status code]

        Args:
            url ([type]): [description]
            status_code ([type]): [description]

        Raises:
            CatsRequestSessionError: [description]

        Returns:
            [type]: [description]
        """
        if status_code != 200:
            raise CatsWalkRequestError(f"{url} response code is {status_code}")
        return True

    def retry_get(self, url, response_content_type=None, retry_num=4, wait=1):
        """[summary]

        Args:
            url ([type]): [description]
            response_content_type ([type], optional): [description]. Defaults to None.
            retry_num (int, optional): [description]. Defaults to 4.
            wait (int, optional): [description]. Defaults to 1.

        Raises:
            CatsRequestSessionError: [description]

        Returns:
            [type]: [description]
        """
        response = None
        for i in range(retry_num):
            try:
                response = self.get(url, response_content_type)
                return response
            except Exception:
                print(f"retry_get: {url} retry {i}")
                time.sleep(wait)
        raise CatsWalkRequestError(f"{url} retry count is {retry_num}")

    def get(self, url, response_content_type=None):
        """[summary]
        
        Arguments:
            url {[type]} -- [description]
        
        Keyword Arguments:
            response_content_type {[type]} -- [html or json] (default: {None})
        
        Raises:
            RuntimeError: [description]
        
        Returns:
            [type] -- [description]
        """
        ret = self.session.get(url, timeout=self.timeout)
        self.__check_status_code(url, ret.status_code)
        return self.__mk_result(ret, response_content_type)

    def post(self, url, post_data, response_content_type):
        """[summary]

        Args:
            url ([type]): [description]
            post_data ([type]): [description]
            response_content_type ([type]): [description]

        Returns:
            [type]: [description]
        """
        ret = self.session.post(url, post_data, timeout=self.timeout)
        self.__check_status_code(url, ret.status_code)
        return self.__mk_result(ret, response_content_type)

    def download(self, url: str, fullpath: str, request_type: str = "get", post_data=None):
        """[summary]

        Args:
            url (str): [description]
            fullpath (str): [description]
            request_type (str, optional): [description]. Defaults to "get".
            post_data ([type], optional): [description]. Defaults to None.
        """
        if request_type == "post":
            bi = self.post(url=url, post_data=post_data)
        else:
            bi = self.get(url=url).content
        with open(fullpath, "wb") as f:
            f.write(bi)

    def get_global_info(self):
        """[summary]

        Returns:
            [type]: [description]
        """
        url = "https://ipinfo.io"
        result = None
        try:
            result = self.get(url=url, response_content_type="json").content
        except Exception:
            logger.warning(f"get_global_info response is null: {sys.exc_info()}")
        return result

    def print_global_info(self):
        """[summary]
        """
        logger.info(f"global network info is {self.get_global_info()}")



import logging
from bs4 import BeautifulSoup
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.by import By
from selenium import webdriver
from selenium.webdriver.chrome.options import Options
from catswalk.scraping.request import CWRequest
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
import os
from catswalk.scraping.types.type_response import ResponseHtml 
from catswalk.scraping.types.type_webdriver import EXECUTION_ENV, DEVICE, DEVICE_MODE
from selenium.webdriver.common.action_chains import ActionChains

class CWWebDriver:
    def __init__(self, binary_location: str = None, executable_path: str = None, execution_env: EXECUTION_ENV = EXECUTION_ENV.LOCAL, device = DEVICE.DESKTOP_GENERAL, proxy: str = None, implicitly_wait = 5.0):
        """[summary]

        Args:
            binary_location (str): [description]
            executable_path (str): [description]
            execution_env (str, optional): [local, local_headless, aws]. Defaults to "local".
            proxy (str, optional): [description]. Defaults to None.
        """
        self.binary_location = binary_location
        self.executable_path = executable_path
        self.execution_env = execution_env
        self.proxy = proxy

        options = Options()
        if self.execution_env == EXECUTION_ENV.LOCAL_HEADLESS:
            options.binary_location = self.binary_location
            options.add_argument('--headless') # https://www.ytyng.com/blog/ubuntu-chromedriver/
            options.add_argument("--disable-dev-shm-usage")  # overcome limited resource problems
            options.add_argument("start-maximized")  # open Browser in maximized mode
            options.add_argument("disable-infobars")  # disabling infobars
            options.add_argument("--disable-extensions")  # disabling extensions
            options.add_argument("--disable-gpu")  # applicable to windows os only
            options.add_argument("--no-sandbox")  # Bypass OS security model
        elif self.execution_env == EXECUTION_ENV.AWS_LAMBDA:
            os.environ['HOME'] = '/opt/browser/'
            self.executable_path = "/opt/browser/chromedriver"
            options.binary_location = "/opt/browser/headless-chromium"
            options.add_argument("--headless")
            options.add_argument("--no-sandbox")
            options.add_argument("--single-process")
            options.add_argument("--disable-gpu")
            options.add_argument("--window-size=1280x1696")
            options.add_argument("--disable-application-cache")
            options.add_argument("--disable-infobars")
            options.add_argument("--hide-scrollbars")
            options.add_argument("--enable-logging")
            options.add_argument("--log-level=0")
            options.add_argument("--ignore-certificate-errors")
            options.add_argument("--homedir=/tmp")
            options.add_argument('--disable-dev-shm-usage')
        else:
            options.binary_location = self.binary_location
        if self.proxy:
            logging.info("WebDriverSession proxy on")
            options.add_argument(f"proxy-server={self.proxy}")

        if device.value.mode == DEVICE_MODE.MOBILE:
            mobile_emulation = { "deviceName": "Galaxy S5" }
            options.add_experimental_option("mobileEmulation", mobile_emulation)

        caps = DesiredCapabilities.CHROME
        caps['loggingPrefs'] = {'performance': 'INFO'}
        #logging.info(f"WebDriverSession.__init__ : {binary_location}, {executable_path}, {proxy}, {execution_env}")
        self.driver = webdriver.Chrome(options=options, executable_path=self.executable_path, desired_capabilities=caps)
        self.driver.implicitly_wait(implicitly_wait)



    def __enter__(self):
        """

        :return:
        """
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """

        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :return:
        """
        self.close()

    def close(self):
        """[Close WebDriverSession, if chromewebdriver dosen't kill, plsease execute "killall chromedriver"]
        
        """
        self.driver.quit()

    def reload(self):
        self.driver.refresh()

    @property
    def cookies(self):
        """[Get cookie info]

        Returns:
            [type]: [description]
        """
        return self.driver.get_cookies()

    def to_request_session(self) -> CWRequest:
        """[summary]
        
        Returns:
            Request -- [description]
        """
        session = CWRequest()
        for cookie in self.driver.get_cookies():
            self.session.cookies.set(cookie["name"], cookie["value"])
        return session

    def wait_rendering_by_id(self, id, timeout=20):
        """[summary]
        
        Arguments:
            id {[type]} -- [description]
        
        Keyword Arguments:
            timeout {int} -- [description] (default: {20})
        """
        WebDriverWait(self.driver, timeout).until(EC.presence_of_element_located((By.ID, id)))

    def wait_rendering_by_class(self, _class, timeout=20):
        """[summary]
        
        Arguments:
            _class {[type]} -- [description]
        
        Keyword Arguments:
            timeout {int} -- [description] (default: {20})
        """
        WebDriverWait(self.driver, 20).until(EC.presence_of_element_located((By.CLASS_NAME, _class)))

    def transition(self, url: str):
        """[summary]

        Args:
            url (str): [description]
        """
        self.driver.get(url)

    def get(self, url: str):
        self.transition(url=url)
        #WebDriverWait(self.driver, 10).until(EC.url_changes(url))
        soup = self.html
        return soup

    def print_screen_by_position(self, w, h, output_path: str, filename:str):
        """[summary]

        Args:
            w ([type]): [description]
            h ([type]): [description]
            output_path (str): [description]
            filename (str): [description]
        """
        # set window size
        self.driver.set_window_size(w, h)

        # Get Screen Shot
        fullpath = f"{output_path}/{filename}.png"
        self.driver.save_screenshot(fullpath)

    def print_screen_by_xpath(self, xpath:str, output_path: str, filename:str) -> str:
        """[summary]

        Args:
            xpath (str): [description]
            output_path (str): [description]
            filename (str): [description]
        """
        # Get Screen Shot
        fullpath = f"{output_path}/{filename}.png"
        # 範囲を指定してスクリーンショットを撮る
        png = self.driver.find_element_by_xpath(xpath).screenshot_as_png
        # ファイルに保存
        with open(fullpath, 'wb') as f:
            f.write(png)
        return fullpath

    def print_screen_by_class_name(self, class_name:str, output_path: str, filename:str) -> str:
        """[summary]

        Args:
            class_name (str): [description]
            output_path (str): [description]
            filename (str): [description]
        """
        # Get Screen Shot
        fullpath = f"{output_path}/{filename}.png"
        # 範囲を指定してスクリーンショットを撮る
        png = self.driver.find_element_by_class_name(class_name).screenshot_as_png
        # ファイルに保存
        with open(fullpath, 'wb') as f:
            f.write(png)
        return fullpath
    
    def print_fullscreen(self, path, filename):
        # get width and height of the page
        w = self.driver.execute_script("return document.body.scrollWidth;")
        h = self.driver.execute_script("return document.body.scrollHeight;")
        self.print_screen_by_position(w, h, path, filename)


    def click_by_class_name(self, class_name:str) -> str:
        """[summary]

        Args:
            class_name (str): [description]

        Returns:
            str: [description]
        """
        # Get Screen Shot
        elem = self.driver.find_element_by_class_name(class_name)
        elem.click()


    def move_to_element_by_class_name(self, class_name:str) -> str:
        """[summary]

        Args:
            class_name (str): [description]

        Returns:
            str: [description]
        """
        # Get Screen Shot
        elem = self.driver.find_element_by_class_name(class_name)
        elem.location_once_scrolled_into_view


    @property
    def html(self):
        html = self.driver.page_source.encode('utf-8')
        return BeautifulSoup(html, "lxml")

    @property
    def log(self):
        result = []
        for entry in self.driver.get_log('performance'):
            result.append(entry['message'])
        return result



import unittest
from catswalk.scraping.webdriver import CWWebDriver
import time
from catswalk.scraping.types.type_webdriver import *

class TestCWWebDriver(unittest.TestCase):
    binary_location = "/Applications/Google Chrome.app/Contents/MacOS/Google Chrome"
    executable_path = "/Users/rv/ws/tools/chromedriver"
    proxy = None
    headless = True
    
    def test_simple_get(self):
        """[python -m unittest tests.scraping.test_web_driver.TestCWWebDriver.test_simple_get]
        """
        """
        request = CWWebDriver(binary_location=self.binary_location, executable_path=self.executable_path)
        html = request.move(url="https://yahoo.co.jp")
        time.sleep(10)
        request.close()
        """
        
        with CWWebDriver(binary_location=self.binary_location, executable_path=self.executable_path) as request:
            responseHtml = request.move(url="https://yahoo.co.jp")
            print(responseHtml)
            print("end")
            time.sleep(10)
        

    def test_print_screen(self):
        """[python -m unittest tests.scraping.test_web_driver.TestCWWebDriver.test_print_screen]
        """
        request = CWWebDriver(binary_location=self.binary_location, executable_path=self.executable_path, execution_env=EXECUTION_ENV.LOCAL,  device = DEVICE.MOBILE_GALAXY_S5)
        request.move(url="https://yahoo.co.jp")
        request.print_fullscreen("/tmp", "test_print_screen")
        time.sleep(10)
        request.close()

    def test_print_screen_by_xpath(self):
        """[python -m unittest tests.scraping.test_web_driver.TestCWWebDriver.test_print_screen_by_xpath]
        """
        request = CWWebDriver(binary_location=self.binary_location, executable_path=self.executable_path)
        request.move(url="https://yahoo.co.jp")
        # xpath:str, output_path: str, filename:str
        request.print_screen_by_xpath(xpath = '//*[@id="ContentWrapper"]/main/div[3]/article[1]', output_path="/tmp", filename="test_print_screen_by_xpath")
        time.sleep(10)
        request.close()

if __name__ == "__main__":
    unittest.main()


from enum import Enum
from dataclasses import dataclass
from typing import List, Mapping

class EXECUTION_ENV(Enum):
    LOCAL = "local"
    LOCAL_HEADLESS = "local_headless"
    AWS_LAMBDA = "aws_lambda"


@dataclass
class DeviceType:
    """[summary]
    """
    mode: str
    agent: str

class DEVICE_MODE(Enum):
    DESKTOP = "Desktop"
    MOBILE = "Mobile"

class DEVICE(Enum):
    DESKTOP_GENERAL = DeviceType(mode=DEVICE_MODE.DESKTOP, agent="general")
    MOBILE_GALAXY_S5 = DeviceType(mode=DEVICE_MODE.MOBILE, agent="Galaxy S5")
    